use std::collections::HashMap;

use axum::http::StatusCode;
use futures::TryStreamExt as _;
use uuid::Uuid;

use crate::Pool;

#[derive(sqlx::Type, Debug)]
#[sqlx(type_name = "permission")]
#[sqlx(rename_all = "lowercase")]
pub enum PermissionRaw {
    Read = 1,
    Write,
    Manage,
}

#[derive(Clone, Copy, Debug, PartialEq, Eq, PartialOrd, Ord, Default)]
pub enum PermissionType {
    #[default]
    NoPermission = 1,
    Read,
    Write,
    Manage,
}

impl From<Option<PermissionRaw>> for PermissionType {
    fn from(value: Option<PermissionRaw>) -> PermissionType {
        use PermissionRaw as PR;
        match value {
            Some(PR::Read) => PermissionType::Read,
            Some(PR::Write) => PermissionType::Write,
            Some(PR::Manage) => PermissionType::Manage,
            None => PermissionType::NoPermission,
        }
    }
}

impl From<PermissionType> for PermissionRaw {
    fn from(value: PermissionType) -> Self {
        match value {
            PermissionType::Manage => Self::Manage,
            PermissionType::Write => Self::Write,
            PermissionType::Read => Self::Read,
            PermissionType::NoPermission => unreachable!(),
        }
    }
}

impl PermissionType {
    pub fn can_read(self) -> bool {
        self >= PermissionType::Read
    }

    pub fn can_read_guard(self) -> Result<(), StatusCode> {
        if !self.can_read() {
            return Err(StatusCode::NOT_FOUND);
        }
        Ok(())
    }

    pub fn can_write_guard(self) -> Result<(), StatusCode> {
        self.can_read_guard()?;
        if self < PermissionType::Write {
            return Err(StatusCode::FORBIDDEN);
        }
        Ok(())
    }

    pub fn can_manage_guard(self) -> Result<(), StatusCode> {
        self.can_read_guard()?;
        if self < PermissionType::Manage {
            return Err(StatusCode::FORBIDDEN);
        }
        Ok(())
    }
}

pub async fn insert_permissions(
    user_id: i32,
    folder_id: Uuid,
    permission_type: PermissionType,
    pool: &Pool,
) -> sqlx::Result<()> {
    let permission_type = PermissionRaw::from(permission_type);
    sqlx::query_file!(
        "sql/insert_permission.sql",
        folder_id,
        user_id,
        permission_type as PermissionRaw
    )
    .execute(pool)
    .await
    .map(|_| ())
}

pub async fn get_permissions_for_folder(
    folder_id: Uuid,
    pool: &Pool,
) -> sqlx::Result<HashMap<String, PermissionType>> {
    sqlx::query_file!("sql/get_all_permissions_for_folder.sql", folder_id)
        .fetch(pool)
        .map_ok(|record| (record.username, Some(record.permission_type).into()))
        .try_collect()
        .await
}
